#ifndef SQL_CONF_H
#define SQL_CONF_H
/***************************************************************************
*  conf.h                             rlm_nibs - FreeRADIUS SQL Module     *
*                                                                          *
*      Configuration options for the SQL module                            *
*                                                                          *
*                                     Mike Machado <mike@innercite.com>    *
*                               Neon aka Sergey Derkach <neon@is.kh.ua>    *
***************************************************************************/

#include "nibs.h"

typedef struct nibs_config {
	char   *nibs_driver;
	char   *nibs_server;
	char   *nibs_port;
	char   *nibs_login;
	char   *nibs_password;
	char   *nibs_db;
	char   *nibs_auth_table;
	char   *nibs_packets_table;
	char   *nibs_prices_table;
	char   *nibs_acct_table;
	char   *nibs_blacklist_table;
	char   *nibs_holidays_table;
	int     sqltrace;
	char   *tracefile;
	int     num_socks;
	int     connect_failure_retry;
	int     connect_failure_retry_delay;
	char   *sql_user_name;
	char   *group_select_query;
	char   *holyday_select_query;
	char   *prices_select_query;
	char   *blacklist_select_query;
	char   *authenticate_query;
	char   *authorize_query;
	char   *authorize_query_prof;
	char   *month_time_query;
	char   *week_time_query;
	char   *day_time_query;
	char   *month_traffic_query;
	char   *week_traffic_query;
	char   *day_traffic_query;
	char   *month_money_query;
	char   *week_money_query;
	char   *day_money_query;
	char   *activation_update_query;
	char   *fixed_select_query;
	char   *statistic_select_query;
	char   *accounting_start_query;
	char   *accounting_start_query_alt;
	char   *accounting_reopen;
	char   *accounting_alive_auth_query;
	char   *accounting_stop_auth_query;
	char   *accounting_update_query;
	char   *accounting_stop_query;
	char   *accounting_onoff_query;
	char   *simul_global_query;
	char   *simul_count_query;
	char   *simul_verify_query;
	int     deletestalesessions_by_checkrad;
	int     deletestalesessions_by_alive;
	int	alive_max_interval;
	int     set_session_timeout;
	int     use_black_list;
	int     kill_by_alive;
	char   *kill_program;
	char   *xlat_name;
	PACKET          *first_packet;
	int              packets;
	HOLIDAY         *first_holiday;
	int              holidays;
	PRICES          *first_price;
	int	         num_prices;

	/* individual driver config */
	void	*localcfg;

} NIBS_CONFIG;


#define CHECKRAD1		"/usr/sbin/checkrad"
#define CHECKRAD2		"/usr/local/sbin/checkrad"

/* Hack for funky ascend ports on MAX 4048 (and probably others)
   The "NAS-Port-Id" value is "xyyzz" where "x" = 1 for digital, 2 for analog;
   "yy" = line number (1 for first PRI/T1/E1, 2 for second, so on);
   "zz" = channel number (on the PRI or Channelized T1/E1).
    This should work with normal terminal servers, unless you have a TS with
        more than 9999 ports ;^).
    The "ASCEND_CHANNELS_PER_LINE" is the number of channels for each line into
        the unit.  For my US/PRI that's 23.  A US/T1 would be 24, and a
        European E1 would be 30 (I think ... never had one ;^).
    This will NOT change the "NAS-Port-Id" reported in the detail log.  This
        is simply to fix the dynamic IP assignments a la Cistron.
    WARNING: This hack works for me, but I only have one PRI!!!  I've not
        tested it on 2 or more (or with models other than the Max 4048)
    Use at your own risk!
  -- dgreer@austintx.com
*/

#define ASCEND_PORT_HACK
#define ASCEND_CHANNELS_PER_LINE        23
#define CISCO_ACCOUNTING_HACK

/* SQL defines */
#define MAX_QUERY_LEN			4096
#define SQL_LOCK_LEN			MAX_QUERY_LEN
#define	SQLTRACEFILE			RADLOG_DIR "/sqltrace.sql"

/* SQL Errors */
#define SQL_DOWN			1 /* for re-connect */

#define MAX_COMMUNITY_LEN		50
#define MAX_SQL_SOCKS			256
#define MAX_TABLE_LEN			20
#define MAX_AUTH_QUERY_LEN		256
#define AUTH_STRING_LEN			128

#endif /* SQL_CONF_H */
