#
#        FreeNIBS  PostgreSQL.
#

#
#  freenibs  :
# su pgsql
# cd /usr/local/bin
# ./createdb --encoding='SQL_ASCII' freenibs
#

#
# 
#

CREATE SEQUENCE public.blacklist_id_seq INCREMENT 1 MINVALUE 1 MAXVALUE 2147483647 CACHE 1;
CREATE SEQUENCE public.packets_gid_seq INCREMENT 1 MINVALUE 1 MAXVALUE 2147483647 CACHE 1;
CREATE SEQUENCE public.user_uid_seq INCREMENT 1 MINVALUE 1 MAXVALUE 2147483647 CACHE 1;

#
#  ,     FreeNIBS.
#
#        plpgsql:
#
# su pgsql
# cd /usr/local/bin
# ./createlang --dbname=freenibs plpgsql
#

CREATE FUNCTION public.from_days(int4) RETURNS timestamp AS '
    select ''0000-01-02''::timestamp + ($1 || '' days'')::interval as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.from_unixtime(int4) RETURNS timestamp AS '
    select abstime($1)::timestamp as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.if(int4, int4, int4) RETURNS int4 AS '
DECLARE
 expr1 ALIAS FOR $1;
 expr2 ALIAS FOR $2;
 expr3 ALIAS FOR $3;

BEGIN
 IF expr1 is null THEN
  return expr3;
 END IF;
 
 IF expr1=0 THEN
  return expr3;
 END IF;

 RETURN expr2;

END;
'  LANGUAGE 'sql' VOLATILE;


CREATE FUNCTION public.if(text, text, text) RETURNS text AS '
DECLARE
 expr1 ALIAS FOR $1;
 expr2 ALIAS FOR $2;
 expr3 ALIAS FOR $3;

BEGIN
 IF expr1 is null THEN
  return expr3;
 END IF;
 
 RETURN expr2;

END;
'  LANGUAGE 'sql' VOLATILE;


CREATE FUNCTION public.if(int8, int8, int8) RETURNS int8 AS '
DECLARE
 expr1 ALIAS FOR $1;
 expr2 ALIAS FOR $2;
 expr3 ALIAS FOR $3;

BEGIN
 IF expr1 is null THEN
  return expr3;
 END IF;
 
 IF expr1=0 THEN
  return expr3;
 END IF;

 RETURN expr2;

END;
'  LANGUAGE 'sql' VOLATILE;
       
CREATE FUNCTION public.ifnull(text, text) RETURNS text AS '
    select coalesce($1, $2) as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.ifnull(int4, int4) RETURNS int4 AS '
    select coalesce($1, $2) as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public."month"(timestamp) RETURNS int4 AS '
    select extract(month from $1)::int4 as result
'  LANGUAGE 'sql' VOLATILE;
       
CREATE FUNCTION public.now() RETURNS timestamp AS '
    select LOCALTIMESTAMP as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.sec_to_time() RETURNS int4 AS '
    select 0::int4 as result
'  LANGUAGE 'sql' VOLATILE;
 
CREATE FUNCTION public.sec_to_time(int4) RETURNS reltime AS '
    select reltime($1)::reltime as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.to_days(timestamp) RETURNS int4 AS '
    select date_part(''day'', $1 - ''0000-01-01'')::int4 as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.to_timestamp(int4) RETURNS timestamp AS '
    select null::timestamp as result
'  LANGUAGE 'sql' VOLATILE;
       
CREATE FUNCTION public.to_timestamp(date) RETURNS date AS '
    select $1::date as result
'  LANGUAGE 'sql' VOLATILE;
       
CREATE FUNCTION public.to_timestamp(text) RETURNS date AS '
    select $1::date as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.unix_timestamp(timestamp) RETURNS int4 AS '
DECLARE
 ts ALIAS FOR $1;
 ret_value int4;
 new_year int4;

BEGIN
 IF ts=''0001-01-01 BC'' THEN
  RETURN 0;
 END IF;

 select into new_year date_part(''YEAR'', ts);

 IF new_year  2020 THEN
  RETURN 0;
 END IF;

 select into ret_value date_part(''epoch'', ts);

 RETURN ret_value;

END;
'  LANGUAGE 'sql' VOLATILE;
	      
CREATE FUNCTION public.unix_timestamp() RETURNS int4 AS '
    select date_part(''epoch'', now())::int4 as result
'  LANGUAGE 'sql' VOLATILE;

CREATE FUNCTION public.week(timestamp) RETURNS int4 AS '
    select extract(week from $1)::int4 as result
'  LANGUAGE 'sql' VOLATILE;
       
CREATE FUNCTION public."year"(timestamp) RETURNS int4 AS '
    select extract(year from $1)::int4 as result
'  LANGUAGE 'sql' VOLATILE;

#
#  
#

CREATE TABLE public.users (
    username varchar(64) binary DEFAULT '' NOT NULL, 
    userpass varchar(254) binary DEFAULT '*' NOT NULL, 
    crypt_method int2 DEFAULT '1' NOT NULL, 
    uid int8 DEFAULT nextval('user_uid_seq'::text) NOT NULL, 
    gid int4 DEFAULT '1' NOT NULL, 
    deposit float8 DEFAULT '0.000000' NOT NULL, 
    credit float8 DEFAULT '0.000000' NOT NULL, 
    fio varchar(128) DEFAULT '' NOT NULL, 
    phone varchar(128) DEFAULT '' NOT NULL, 
    address varchar(128) DEFAULT '' NOT NULL, 
    prim varchar(254) DEFAULT '' NOT NULL, 
    add_date date DEFAULT '0000-00-00' NOT NULL, 
    blocked int2 DEFAULT '0' NOT NULL, 
    activated int2 DEFAULT '1' NOT NULL, 
    expired date NOT NULL default '0000-00-00', 
    total_time int4 DEFAULT 0 NOT NULL, 
    total_traffic int8 DEFAULT '0' NOT NULL, 
    total_money float8 DEFAULT '0.000000' NOT NULL, 
    last_connection date DEFAULT '0000-00-00' NOT NULL, 
    framed_ip varchar(16) DEFAULT '' NOT NULL, 
    framed_mask varchar(16) DEFAULT '' NOT NULL, 
    callback_number varchar(64) DEFAULT '' NOT NULL, 
    CONSTRAINT users_pkey PRIMARY KEY (uid)
) WITH OIDS;

CREATE TABLE public.actions (
    username varchar(64) DEFAULT '' NOT NULL, 
    gid int4 DEFAULT '0' NOT NULL,
    id varchar(32) DEFAULT '' NOT NULL, 
    unique_id varchar(64) DEFAULT '' NOT NULL, 
    time_on int4, 
    start_time timestamp DEFAULT '2003-01-01 00:00:00', 
    stop_time timestamp DEFAULT '0000-00-00 00:00:00', 
    in_bytes int8, 
    out_bytes int8, 
    ip varchar(16) DEFAULT '' NOT NULL, 
    server varchar(16) DEFAULT '' NOT NULL, 
    client_ip varchar(16) DEFAULT '' NOT NULL, 
    port int4, 
    call_to varchar(64) DEFAULT '' NOT NULL, 
    call_from varchar(64) DEFAULT '' NOT NULL, 
    connect_info varchar(254), 
    protocol varchar(32),
    terminate_cause varchar(64) DEFAULT '' NOT NULL, 
    hour_in_bytes int8 DEFAULT '0' NOT NULL,
    hour_out_bytes int8 DEFAULT '0' NOT NULL,
    hour_traffic_money float8 DEFAULT '0.000000' NOT NULL, 
    last_change int4 DEFAULT '0' NOT NULL,
    before_billing float8 DEFAULT '0.000000' NOT NULL, 
    billing_minus float8 DEFAULT '0.000000' NOT NULL, 
) WITH OIDS;

CREATE TABLE public.packets (
    num int4 DEFAULT '0' NOT NULL, 
    packet varchar(128) DEFAULT '' NOT NULL, 
    prefix varchar(4) DEFAULT '' NOT NULL, 
    gid int4 DEFAULT nextval('packets_gid_seq'::text) NOT NULL, 
    deposit float8 DEFAULT '0.000000' NOT NULL, 
    credit float8 DEFAULT '0.000000' NOT NULL, 
    tos int2 DEFAULT '0' NOT NULL, 
    do_with_tos int2 DEFAULT '1' NOT NULL, 
    direction int2 DEFAULT '0' NOT NULL, 
    fixed int2 DEFAULT '0' NOT NULL, 
    fixed_cost float8 DEFAULT '0.000000' NOT NULL, 
    activated int2 DEFAULT '1' NOT NULL, 
    activation_time int8 DEFAULT '2851200' NOT NULL, 
    blocked int2 DEFAULT '0' NOT NULL, 
    total_time_limit int8 DEFAULT '0' NOT NULL, 
    month_time_limit int8 DEFAULT '0' NOT NULL, 
    week_time_limit int8 DEFAULT '0' NOT NULL, 
    day_time_limit int8 DEFAULT '0' NOT NULL, 
    total_traffic_limit int8 DEFAULT '0' NOT NULL, 
    month_traffic_limit int8 DEFAULT '0' NOT NULL, 
    week_traffic_limit int8 DEFAULT '0' NOT NULL, 
    day_traffic_limit int8 DEFAULT '0' NOT NULL, 
    total_money_limit float8 DEFAULT '0.000000', 
    month_money_limit float8 DEFAULT '0.000000', 
    week_money_limit float8 DEFAULT '0.000000', 
    day_money_limit float8 DEFAULT '0.000000', 
    login_time varchar(254) DEFAULT '' NOT NULL, 
    huntgroup_name varchar(64) DEFAULT '' NOT NULL, 
    simultaneous_use int2 DEFAULT '0' NOT NULL, 
    port_limit int4 DEFAULT '0' NOT NULL, 
    session_timeout int8 DEFAULT '0' NOT NULL, 
    idle_timeout int8 DEFAULT '0' NOT NULL, 
    allowed_prefixes varchar(64) DEFAULT '' NOT NULL, 
    framed_ip varchar(16) DEFAULT '' NOT NULL, 
    framed_mask varchar(16) DEFAULT '' NOT NULL, 
    no_pass int2 DEFAULT '0' NOT NULL, 
    no_acct int2 DEFAULT '0' NOT NULL, 
    allow_callback int2 DEFAULT '0' NOT NULL, 
    other_params varchar(254) DEFAULT 'Acct-Interim-Interval = 60, Service-Type = Framed-User, Framed-Protocol = PPP, Framed-Compression = Van-Jacobson-TCP-IP' NOT NULL, 
    create_system_user int2 DEFAULT '0' NOT NULL, 
    CONSTRAINT packets_pkey PRIMARY KEY (gid)
) WITH OIDS;

CREATE TABLE public.prices (
    gid int4 DEFAULT '1' NOT NULL, 
    week_day int2 DEFAULT '0' NOT NULL, 
    h0 float8 DEFAULT '0.000000' NOT NULL, 
    input0 float8 DEFAULT '0.000000' NOT NULL, 
    output0 float8 DEFAULT '0.000000' NOT NULL, 
    h1 float8 DEFAULT '0.000000' NOT NULL, 
    input1 float8 DEFAULT '0.000000' NOT NULL, 
    output1 float8 DEFAULT '0.000000' NOT NULL, 
    h2 float8 DEFAULT '0.000000' NOT NULL, 
    input2 float8 DEFAULT '0.000000' NOT NULL, 
    output2 float8 DEFAULT '0.000000' NOT NULL, 
    h3 float8 DEFAULT '0.000000' NOT NULL, 
    input3 float8 DEFAULT '0.000000' NOT NULL, 
    output3 float8 DEFAULT '0.000000' NOT NULL, 
    h4 float8 DEFAULT '0.000000' NOT NULL, 
    input4 float8 DEFAULT '0.000000' NOT NULL, 
    output4 float8 DEFAULT '0.000000' NOT NULL, 
    h5 float8 DEFAULT '0.000000' NOT NULL, 
    input5 float8 DEFAULT '0.000000' NOT NULL, 
    output5 float8 DEFAULT '0.000000' NOT NULL, 
    h6 float8 DEFAULT '0.000000' NOT NULL, 
    input6 float8 DEFAULT '0.000000' NOT NULL, 
    output6 float8 DEFAULT '0.000000' NOT NULL, 
    h7 float8 DEFAULT '0.000000' NOT NULL, 
    input7 float8 DEFAULT '0.000000' NOT NULL, 
    output7 float8 DEFAULT '0.000000' NOT NULL, 
    h8 float8 DEFAULT '0.000000' NOT NULL, 
    input8 float8 DEFAULT '0.000000' NOT NULL, 
    output8 float8 DEFAULT '0.000000' NOT NULL, 
    h9 float8 DEFAULT '0.000000' NOT NULL, 
    input9 float8 DEFAULT '0.000000' NOT NULL, 
    output9 float8 DEFAULT '0.000000' NOT NULL, 
    h10 float8 DEFAULT '0.000000' NOT NULL, 
    input10 float8 DEFAULT '0.000000' NOT NULL, 
    output10 float8 DEFAULT '0.000000' NOT NULL, 
    h11 float8 DEFAULT '0.000000' NOT NULL, 
    input11 float8 DEFAULT '0.000000' NOT NULL, 
    output11 float8 DEFAULT '0.000000' NOT NULL, 
    h12 float8 DEFAULT '0.000000' NOT NULL, 
    input12 float8 DEFAULT '0.000000' NOT NULL, 
    output12 float8 DEFAULT '0.000000' NOT NULL, 
    h13 float8 DEFAULT '0.000000' NOT NULL, 
    input13 float8 DEFAULT '0.000000' NOT NULL, 
    output13 float8 DEFAULT '0.000000' NOT NULL, 
    h14 float8 DEFAULT '0.000000' NOT NULL, 
    input14 float8 DEFAULT '0.000000' NOT NULL, 
    output14 float8 DEFAULT '0.000000' NOT NULL, 
    h15 float8 DEFAULT '0.000000' NOT NULL, 
    input15 float8 DEFAULT '0.000000' NOT NULL, 
    output15 float8 DEFAULT '0.000000' NOT NULL, 
    h16 float8 DEFAULT '0.000000' NOT NULL, 
    input16 float8 DEFAULT '0.000000' NOT NULL, 
    output16 float8 DEFAULT '0.000000' NOT NULL, 
    h17 float8 DEFAULT '0.000000' NOT NULL, 
    input17 float8 DEFAULT '0.000000' NOT NULL, 
    output17 float8 DEFAULT '0.000000' NOT NULL, 
    h18 float8 DEFAULT '0.000000' NOT NULL, 
    input18 float8 DEFAULT '0.000000' NOT NULL, 
    output18 float8 DEFAULT '0.000000' NOT NULL, 
    h19 float8 DEFAULT '0.000000' NOT NULL, 
    input19 float8 DEFAULT '0.000000' NOT NULL, 
    output19 float8 DEFAULT '0.000000' NOT NULL, 
    h20 float8 DEFAULT '0.000000' NOT NULL, 
    input20 float8 DEFAULT '0.000000' NOT NULL, 
    output20 float8 DEFAULT '0.000000' NOT NULL, 
    h21 float8 DEFAULT '0.000000' NOT NULL, 
    input21 float8 DEFAULT '0.000000' NOT NULL, 
    output21 float8 DEFAULT '0.000000' NOT NULL, 
    h22 float8 DEFAULT '0.000000' NOT NULL, 
    input22 float8 DEFAULT '0.000000' NOT NULL, 
    output22 float8 DEFAULT '0.000000' NOT NULL, 
    h23 float8 DEFAULT '0.000000' NOT NULL,
    input23 float8 DEFAULT '0.000000' NOT NULL, 
    output23 float8 DEFAULT '0.000000' NOT NULL, 
) WITH OIDS;

CREATE TABLE public.holidays (
    holiday_date varchar(5) DEFAULT '' NOT NULL, 
    "comment" varchar(64) DEFAULT '' NOT NULL, 
    CONSTRAINT holidays_pkey PRIMARY KEY (holiday_date)
) WITH OIDS;

CREATE TABLE public.blacklist (
    id int4 DEFAULT nextval('blacklist_id_seq'::text) NOT NULL, 
    uid int8, 
    gid int4, 
    phone varchar(64) DEFAULT '' NOT NULL, 
    permit varchar(1) DEFAULT 'Y' NOT NULL, 
    "comment" varchar(128), 
    CONSTRAINT blacklist_pkey PRIMARY KEY (id)
) WITH OIDS;
