/*
 * rlm_nibs.c		SQL Module
 * 		Main SQL module file. Most ICRADIUS code is located in sql.c
 *
 * Version:	$Id: rlm_nibs.c,v 1.108.2.1 2002/12/04 16:59:29 aland Exp $
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright 2000  The FreeRADIUS server project
 * Copyright 2000  Mike Machado <mike@innercite.com>
 * Copyright 2000  Alan DeKok <aland@ox.org>
 * Copyright 2003  Neon aka Sergey Derkach <neon@is.kh.ua>
 */

static const char rcsid[] =
	"$Id: rlm_nibs.c,v 1.108.2.1 2002/12/04 16:59:29 aland Exp $";

#include "autoconf.h"

#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "radiusd.h"
#include "modules.h"
#include "conffile.h"
#include "rlm_nibs.h"
#include "nibs_utils.h"
#include "rad_assert.h"

#include "../../include/md5.h"
#include "../../include/sha1.h"

#define PW_ACCT_INTERIM_INTERVAL	85

static CONF_PARSER module_config[] = {
        {"driver",PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_driver), NULL, "mysql"},
	{"server",PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_server), NULL, "localhost"},
        {"port",PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_port), NULL, ""},
	{"login", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_login), NULL, "freenibs"},
        {"password", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_password), NULL, ""},
        {"nibs_db", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_db), NULL, "freenibs"},
        {"auth_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_auth_table), NULL, "users"},
        {"packets_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_packets_table), NULL, "packets"},
        {"prices_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_prices_table), NULL, "prices"},
        {"acct_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_acct_table), NULL, "actions"},
        {"blacklist_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_blacklist_table), NULL, "blacklist"},
        {"holidays_table", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,nibs_holidays_table), NULL, "holidays"},
        {"sqltrace", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,sqltrace), NULL, "0"},
        {"sqltracefile", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,tracefile), NULL, SQLTRACEFILE},
        {"num_socks", PW_TYPE_INTEGER, offsetof(NIBS_CONFIG,num_socks), NULL, "64"},
        {"connect_failure_retry", PW_TYPE_INTEGER, offsetof(NIBS_CONFIG,connect_failure_retry), NULL, "10"},
        {"connect_failure_retry_delay", PW_TYPE_INTEGER, offsetof(NIBS_CONFIG,connect_failure_retry_delay), NULL, "10"},
        {"sql_user_name", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,sql_user_name), NULL, ""},
        {"group_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,group_select_query), NULL, ""},
        {"holyday_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,holyday_select_query), NULL, ""},
        {"prices_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,prices_select_query), NULL, ""},
        {"blacklist_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,blacklist_select_query), NULL, ""},
        {"authenticate_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,authenticate_query), NULL, ""},
        {"authorize_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,authorize_query), NULL, ""},
        {"authorize_query_prof", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,authorize_query_prof), NULL, ""},
        {"month_time_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,month_time_query), NULL, ""},
        {"week_time_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,week_time_query), NULL, ""},
        {"day_time_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,day_time_query), NULL, ""},
        {"month_traffic_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,month_traffic_query), NULL, ""},
        {"week_traffic_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,week_traffic_query), NULL, ""},
        {"day_traffic_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,day_traffic_query), NULL, ""},
        {"month_money_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,month_money_query), NULL, ""},
        {"week_money_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,week_money_query), NULL, ""},
        {"day_money_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,day_money_query), NULL, ""},
        {"activation_update_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,activation_update_query), NULL, ""},
        {"fixed_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,fixed_select_query), NULL, ""},
        {"statistic_select_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,statistic_select_query), NULL, ""},
        {"accounting_start_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_start_query), NULL, ""},
        {"accounting_start_query_alt", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_start_query_alt), NULL, ""},
        {"accounting_reopen", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_reopen), NULL, ""},
        {"accounting_alive_auth_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_alive_auth_query), NULL, ""},
        {"accounting_stop_auth_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_stop_auth_query), NULL, ""},
        {"accounting_update_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_update_query), NULL, ""},
        {"accounting_stop_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_stop_query), NULL, ""},
        {"accounting_onoff_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,accounting_onoff_query), NULL, ""},
        {"simul_count_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,simul_count_query), NULL, ""},
        {"simul_verify_query", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,simul_verify_query), NULL, ""},
        {"deletestalesessions_by_checkrad", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,deletestalesessions_by_checkrad), NULL, "0"},
        {"deletestalesessions_by_alive", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,deletestalesessions_by_alive), NULL, "0"},
        {"alive_max_interval", PW_TYPE_INTEGER, offsetof(NIBS_CONFIG,alive_max_interval), NULL, "30"},
        {"set_session_timeout", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,set_session_timeout), NULL, "0"},
        {"use_black_list", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,use_black_list), NULL, "1"},
        {"kill_by_alive", PW_TYPE_BOOLEAN, offsetof(NIBS_CONFIG,kill_by_alive), NULL, "1"},
        {"kill_program", PW_TYPE_STRING_PTR, offsetof(NIBS_CONFIG,kill_program), NULL, "/bin/false"},
	{NULL, -1, 0, NULL, NULL}
};

/***********************************************************************
 * start of main routines
 ***********************************************************************/
static int rlm_nibs_init(void) {

	/*
	 * FIXME:
	 * We should put the sqlsocket array here once
	 * the module code is reworked to not unload
	 * modules on HUP.  This way we can have
	 * persistant connections.  -jcarneal
	 */
	return 0;
}

/*
 *	Yucky prototype.
 */
static int sql_set_user(SQL_INST *inst, REQUEST *request, char *sqlusername, const char *username);

/*
 *	sql xlat function. Right now only SELECTs are supported. Only
 *	the first element of the SELECT result will be used.
 */
static int sql_xlat(void *instance, REQUEST *request, char *fmt, char *out, int freespace,
			RADIUS_ESCAPE_STRING func)
{
	SQLSOCK *sqlsocket;
	SQL_ROW row;
	SQL_INST *inst = instance;
	char querystr[MAX_QUERY_LEN];
	char sqlusername[2 * MAX_STRING_LEN + 10];
	int ret = 0;

	DEBUG("rlm_nibs (%s): - sql_xlat", inst->config->xlat_name);

	/*
         * Add SQL-User-Name attribute just in case it is needed
         *  We could search the string fmt for SQL-User-Name to see if this is
         *  needed or not
         */
	sql_set_user(inst, request, sqlusername, NULL);
	/*
	 * Do an xlat on the provided string (nice recursive operation).
	 */
	if (!radius_xlat(querystr, sizeof(querystr), fmt, request, func)) {
		radlog(L_ERR, "rlm_nibs (sql_xlat): xlat failed.");
		return 0;
	}

	query_log(request, inst, querystr);
	sqlsocket = sql_get_socket(inst);
	if (sqlsocket == NULL)
		return 0;
	if (rlm_nibs_select_query(sqlsocket,inst,querystr)){
		radlog(L_ERR, "rlm_nibs (sql_xlat): database query error, `%s': `%s'",
		       querystr,
		       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
		sql_release_socket(inst,sqlsocket);
		return 0;
	}

	ret = rlm_nibs_fetch_row(sqlsocket, inst);

	if (ret) {
		DEBUG("rlm_nibs (sql_xlat): SQL query did not succeed");
		(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
		sql_release_socket(inst,sqlsocket);
		return 0;
	}

	row = sqlsocket->row;
	if (row == NULL) {
		DEBUG("rlm_nibs (sql_xlat): SQL query did not return any results");
		(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
		sql_release_socket(inst,sqlsocket);
		return 0;
	}

	if (row[0] == NULL){
		DEBUG("rlm_nibs (sql_xlat): row[0] returned NULL");
		(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
		sql_release_socket(inst,sqlsocket);
		return 0;
	}
	ret = strlen(row[0]);
	if (ret > freespace){
		DEBUG("rlm_nibs (sql_xlat): sql_xlat:: Insufficient string space");
		(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
		sql_release_socket(inst,sqlsocket);
		return 0;
	}

	strncpy(out,row[0],ret);

	DEBUG("rlm_nibs (sql_xlat): - sql_xlat finished");

	(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
	sql_release_socket(inst,sqlsocket);
	return ret;
}

/*
 *	Translate the SQL queries.
 */
int sql_escape_func(char *out, int outlen, const char *in)
{
	int len = 0;
	
	while (in[0]) {
		/*
		 *  Only one byte left.
		 */
		if (outlen <= 1) {
			break;
		}
		
		/*
		 *	Non-printable characters get replaced with their
		 *	mime-encoded equivalents.
		 */
		if ((in[0] < 32) ||
		    strchr("@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.-_: =/", *in) == NULL) {
			snprintf(out, outlen, "=%02X", (unsigned char) in[0]);
			in++;
			out += 3;
			outlen -= 3;
			len += 3;
			continue;
		}
		
		/*
		 *	Else it's a nice character.
		 */
		*out = *in;
		out++;
		in++;
		outlen--;
		len++;
	}
	*out = '\0';
	return len;
}

/*
 *	Set the SQL user name.
 */
static int sql_set_user(SQL_INST *inst, REQUEST *request, char *sqlusername, const char *username)
{
	VALUE_PAIR *vp=NULL;
	char tmpuser[MAX_STRING_LEN];

	tmpuser[0]=0;
	sqlusername[0]=0;

	/* Remove any user attr we added previously */
	pairdelete(&request->packet->vps, PW_SQL_USER_NAME);

	if (username != NULL) {
		strNcpy(tmpuser, username, MAX_STRING_LEN);
	} else if (strlen(inst->config->sql_user_name)) {
		radius_xlat(tmpuser, sizeof(tmpuser), inst->config->sql_user_name, request, sql_escape_func);
	} else {
		return 0;
	}

	if (*tmpuser) {
		strNcpy(sqlusername, tmpuser, sizeof(tmpuser));
		DEBUG2("rlm_nibs (sql_set_user): sql_set_user escaped user --> '%s'",
		       sqlusername);
		vp = pairmake("SQL-User-Name", sqlusername, 0);
		if (vp == NULL) {
			radlog(L_ERR, "rlm_nibs (sql_set_user): %s", librad_errstr);
			return -1;
		}

		pairadd(&request->packet->vps, vp);
		return 0;
	}
	return -1;
}

static int rlm_nibs_detach(void *instance)
{
	SQL_INST *inst = instance;

	if (inst->sqlpool) {
		sql_poolfree(inst);
	}
	if (inst->config->xlat_name) {
		xlat_unregister(inst->config->xlat_name,sql_xlat);
		free(inst->config->xlat_name);
	}

	if (inst->config) {
		int i;

		/*
		 *	Free up dynamically allocated string pointers.
		 */
		for (i = 0; module_config[i].name != NULL; i++) {
			char **p;
			if (module_config[i].type != PW_TYPE_STRING_PTR) {
				continue;
			}
			
			/*
			 *	Treat 'config' as an opaque array of bytes,
			 *	and take the offset into it.  There's a
			 *      (char*) pointer at that offset, and we want
			 *	to point to it.
			 */
			p = (char **) (((char *)inst->config) + module_config[i].offset);
			if (!*p) { /* nothing allocated */
				continue;
			}
			free(*p);
			*p = NULL;
		}
		free(inst->config);
		inst->config = NULL;
	}

	if (inst->handle) {
#if 0
		/*
		 *	FIXME: Call the modules 'destroy' function?
		 */
		lt_dlclose(inst->handle);	/* ignore any errors */
#endif
	}
	free(inst);

	return 0;
}

static int rlm_nibs_instantiate(CONF_SECTION * conf, void **instance)
{
	SQL_INST *inst;
	char *xlat_name;

	inst = rad_malloc(sizeof(SQL_INST));
	memset(inst, 0, sizeof(SQL_INST));

	inst->config = rad_malloc(sizeof(NIBS_CONFIG));
	memset(inst->config, 0, sizeof(NIBS_CONFIG));

	/*
	 * If the configuration parameters can't be parsed, then
	 * fail.
	 */
	if (cf_section_parse(conf, inst->config, module_config) < 0) {
		rlm_nibs_detach(inst);
		return -1;
	}

	xlat_name = cf_section_name2(conf);
	if (xlat_name == NULL)
		xlat_name = cf_section_name1(conf);
	if (xlat_name){
		inst->config->xlat_name = strdup(xlat_name);
		xlat_register(xlat_name, sql_xlat, inst);
	}

	if (inst->config->num_socks > MAX_SQL_SOCKS) {
		radlog(L_ERR | L_CONS, "rlm_nibs (rlm_nibs_instantiate): nibs_instantiate: number of sqlsockets cannot exceed MAX_SQL_SOCKS, %d", MAX_SQL_SOCKS);
		rlm_nibs_detach(inst);
		return -1;
	}

	/*
	 *      Sanity check for crazy people.
	 */
	if (strncmp(inst->config->nibs_driver, "rlm_nibs_", 9) != 0) {
		radlog(L_ERR, "rlm_nibs (rlm_nibs_instantiate): \"%s\" is NOT an SQL driver!",
		       inst->config->nibs_driver);
		rlm_nibs_detach(inst);
		return -1;
	}

	inst->handle = lt_dlopenext(inst->config->nibs_driver);
	if (inst->handle == NULL) {
		radlog(L_ERR, "rlm_nibs (rlm_nibs_instantiate): Could not link driver %s: %s",
		       inst->config->nibs_driver,
		       lt_dlerror());
		radlog(L_ERR, "rlm_nibs (rlm_nibs_instantiate): Make sure it (and all its dependent libraries!) are in the search path of your system's ld.");
		rlm_nibs_detach(inst);
		return -1;
	}

	inst->module = (rlm_nibs_module_t *) lt_dlsym(inst->handle, inst->config->nibs_driver);
	if (!inst->module) {
		radlog(L_ERR, "rlm_nibs (rlm_nibs_instantiate): Could not link symbol %s: %s",
		       inst->config->nibs_driver,
		       lt_dlerror());
		rlm_nibs_detach(inst);
		return -1;
	}

	radlog(L_INFO, "rlm_nibs (rlm_nibs_instantiate): Driver %s (module %s) loaded and linked",
	       inst->config->nibs_driver,
	       inst->module->name);
	radlog(L_INFO, "rlm_nibs (rlm_nibs_instantiate): Attempting to connect to %s@%s:%s/%s",
	       inst->config->nibs_login,
	       inst->config->nibs_server, inst->config->nibs_port,
	       inst->config->nibs_db);

	if (sql_init_socketpool(inst) < 0) {
		rlm_nibs_detach(inst);
		return -1;
	}

	if (nibs_init(inst, NULL) != 0) {
    		radlog(L_ERR, "rlm_nibs (rlm_nibs_instantiate): Could not init NiBS");
		rlm_nibs_detach(inst);
    		return -1;
	}

	*instance = inst;

	return RLM_MODULE_OK;
}

static int rlm_nibs_destroy(void)
{
	return 0;
}

static int rlm_nibs_authorize(void *instance, REQUEST * request)
{
	SQL_INST	*inst = instance;
	VALUE_PAIR	*reply_attr;
	NIBS_USER	*nibs_user;
	int		crypt_type;
	VALUE_PAIR	*vp;
	char		querystr[MAX_QUERY_LEN];
	char		querystr2[MAX_QUERY_LEN];
	char		*call_from;
	/* sqlusername holds the sql escaped username. The original
	 * username is at most MAX_STRING_LEN chars long and
	 * *sql_escape_string doubles its length in the worst case.
	 * Throw in an extra 10 to account for trailing NULs and to have
	 * a safety margin. */
        char   sqlusername[2 * MAX_STRING_LEN + 10];

        if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL) {
	    call_from = vp->strvalue;
	} else {
	    call_from = NULL;
	}

	/*
	 *	They MUST have a user name and password to do NIBS authorize.
	 */
	if ((request->username == NULL) || (request->username->length == 0)) {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authorize): Zero length username not permitted [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_AUTH, querystr);
		return RLM_MODULE_INVALID;
	}

	if (sql_set_user(inst, request, sqlusername, NULL) < 0)
	    return RLM_MODULE_FAIL;

	/*
	 *	User must not have a trailing space in login name
	 */

	if (request->username->strvalue[request->username->length - 1] == ' ') {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authorize): Trailing space at login attempt for user `%{SQL-User-Name}' [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_AUTH, querystr);
		return RLM_MODULE_INVALID;
	}

	reply_attr = pairfind(request->config_items, PW_AUTHTYPE);
	
	if((nibs_user = nibs_fill_user(inst, request)) != NULL) {

//ADD ATTRS
	    if (nibs_add_attrs(nibs_user, inst, request) != 0) {
		free(nibs_user);
		pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
		return RLM_MODULE_REJECT;
	    }

	    crypt_type = nibs_user->crypt_method;
	    free(nibs_user);
    	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);

	    if (reply_attr != NULL) {
		if (strcmp(reply_attr->strvalue,"Nibs") != 0) {
		    return RLM_MODULE_NOOP;
		} else {
    		    pairdelete(&request->config_items, PW_AUTHTYPE);
		    reply_attr = NULL;
		}
	    }
	    if(reply_attr == NULL) {
		/* Set Auth-Type = Nibs if Auth-Type is not set */
		if (crypt_type > 0) {
		    reply_attr = pairmake("Auth-Type", "Nibs", T_OP_SET);
		    rad_assert(reply_attr != NULL);
		    pairadd(&request->config_items, reply_attr);
		}
	    }

	} else {
    	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);

	    if (reply_attr != NULL) {
		if (strcmp(reply_attr->strvalue,"Nibs") == 0) {
		    return RLM_MODULE_REJECT;
		} else {
		    return RLM_MODULE_NOTFOUND;
		}
	    } else {
		return RLM_MODULE_NOTFOUND;
	    }
	}


	if (paircmp(NULL, request->packet->vps, request->config_items, NULL) != 0) {
	    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authorize): Check items do not match with received packet (maybe _open_ passwd or huntgroup) for user `%{SQL-User-Name}' [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
	    snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
	    radlog(L_AUTH, querystr);
	    DEBUG2("rlm_nibs (rlm_nibs_authorize): ======== Pairs DEBUG ========");
	    DEBUG2("rlm_nibs (rlm_nibs_authorize): ---------------- Received Packet");
	    vp_listdebug(request->packet->vps);
	    DEBUG2("rlm_nibs (rlm_nibs_authorize): ---------------- Check Items");
	    vp_listdebug(request->config_items);
#if 0
	    DEBUG2("rlm_nibs (rlm_nibs_authorize): ---------------- Reply Items");
	    vp_listdebug(request->reply->vps);
#endif
	    DEBUG2("rlm_nibs (rlm_nibs_authorize): ======== Pairs DEBUG ========");
	    return RLM_MODULE_REJECT;
	}

	return RLM_MODULE_OK;
}

static int rlm_nibs_authenticate(void *instance, REQUEST * request)
{
	SQLSOCK 	*sqlsocket;
	SQL_INST	*inst = instance;
	SQL_ROW		row;
	PACKET	 	*n_packet;
	char		querystr[MAX_QUERY_LEN];
	char		querystr2[MAX_QUERY_LEN];
	char		*enc_pass;
	MD5_CTX		md5_context;
	SHA1_CTX	sha1_context;
	char		digest[20];
	char		buff[MAX_STRING_LEN];
	VALUE_PAIR	*vp;
	char		*call_from;
	/* sqlusername holds the sql escaped username. The original
	 * username is at most MAX_STRING_LEN chars long and
	 * *sql_escape_string doubles its length in the worst case.
	 * Throw in an extra 10 to account for trailing NULs and to have
	 * a safety margin. */
	char   sqlusername[2 * MAX_STRING_LEN + 10];

        if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL) {
	    call_from = vp->strvalue;
	} else {
	    call_from = NULL;
	}

	/*
	 *	They MUST have a user name and password to do NIBS authentication.
	 */
	if ((request->username == NULL) ||
	    (request->username->length == 0)) {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authenticate): Zero length username not permitted [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_ERR, querystr);
		return RLM_MODULE_INVALID;
	}

	/*
	 * Set, escape, and check the user attr here
	 */
	if (sql_set_user(inst, request, sqlusername, NULL) < 0)
		return RLM_MODULE_FAIL;

	radius_xlat(querystr, sizeof(querystr), inst->config->authenticate_query, request, sql_escape_func);
	if(*querystr) {
	    sqlsocket = sql_get_socket(inst);
	    if (sqlsocket == NULL) {
		    /* Remove the username we (maybe) added above */
		    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
		    return(RLM_MODULE_FAIL);
	    }

	    query_log(request, inst, querystr);
    	    if (rlm_nibs_select_query(sqlsocket, inst, querystr)) {
        	    radlog(L_ERR, "rlm_nibs (rlm_nibs_authenticate): database query error - `%s': `%s'", querystr, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
		    /* Remove the username we (maybe) added above */
		    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
        	    sql_release_socket(inst, sqlsocket);
        	    return RLM_MODULE_FAIL;
    	    }
	} else {
	    radlog(L_ERR, "nibs_rlm (rlm_nibs_authenticate): no `authenticate_query' configured in nibs.conf file");
	    /* Remove the username we (maybe) added above */
	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
    	    return RLM_MODULE_FAIL;
	}

        if ((inst->module->sql_num_rows)(sqlsocket, inst->config) == 0) {
	    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authenticate): User `%{SQL-User-Name}' not found in database [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
	    snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
	    radlog(L_AUTH, querystr);
	    (inst->module->sql_finish_select_query)(sqlsocket, inst->config);
            sql_release_socket(inst, sqlsocket);
            return RLM_MODULE_NOTFOUND;
        }

        if(rlm_nibs_fetch_row(sqlsocket, inst)) {
		DEBUG("rlm_nibs (rlm_nibs_authenticate): SQL query did not succeed for user `%s'", request->username->strvalue);
		(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
		sql_release_socket(inst,sqlsocket);
		return RLM_MODULE_FAIL;
	}

	row = sqlsocket->row;
	n_packet = nibs_packet_find(inst, Atoi(row[2]));

	if(n_packet->no_pass) {
	    (inst->module->sql_finish_select_query)(sqlsocket, inst->config);
            sql_release_socket(inst, sqlsocket);
	    /* Remove the username we (maybe) added above */
	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	    return RLM_MODULE_OK;
	}

	if ((!request->password) || (request->password->length == 0)) {
	    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authenticate): Zero length password not permitted for user `%{SQL-User-Name}' [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
	    snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
	    radlog(L_AUTH, querystr);
	    (inst->module->sql_finish_select_query)(sqlsocket, inst->config);
            sql_release_socket(inst, sqlsocket);
	    /* Remove the username we (maybe) added above */
	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	    return RLM_MODULE_INVALID;
	}

	if ((row[1] == NULL) || (row[1][0] == 0)) {
	    radlog(L_AUTH, "rlm_nibs (rlm_nibs_authenticate): Empty password in sql table for user '%s'", request->username->strvalue);
	    (inst->module->sql_finish_select_query)(sqlsocket, inst->config);
            sql_release_socket(inst, sqlsocket);
	    /* Remove the username we (maybe) added above */
	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	    return RLM_MODULE_REJECT;
	}

	/* 
	 *  Switch encryption type, carrently supported are:
	 *    0 - Clear text
	 *    1 - Crypt()
	 *    2 - MD5()
	 *    3 - SHA1()
	 */
	switch (Atoi(row[3])) {
	    case 0:
		enc_pass = strdup(request->password->strvalue);
		break;
	    case 1:
		enc_pass = strdup(crypt(request->password->strvalue, row[1]));
		break;
	    case 2:
		if (strlen(row[1]) != 32) {
		    DEBUG("rlm_nibs (rlm_nibs_authenticate): Configured MD5 password for user '%s' has incorrect length", request->username->strvalue);
		    return RLM_MODULE_REJECT;
		}

		MD5Init(&md5_context);
		MD5Update(&md5_context, request->password->strvalue, request->password->length);
		MD5Final(digest, &md5_context);
		Pap_hexify(buff,digest,16);
		buff[32] = '\0';

		enc_pass = strdup(buff);
		break;
	    case 3:
		if (strlen(row[1]) != 40) {
		    DEBUG("rlm_nibs (rlm_nibs_authenticate): Configured SHA1 password for user '%s' has incorrect length", request->username->strvalue);
		    return RLM_MODULE_REJECT;
		}
		
		SHA1Init(&sha1_context);
		SHA1Update(&sha1_context, request->password->strvalue, request->password->length);
		SHA1Final(digest,&sha1_context);
		Pap_hexify(buff,digest,20);
		buff[40] = '\0';

		enc_pass = strdup(buff);
		break;
	    default:
		enc_pass = strdup(row[1]);
		break;
	}

	if (strcmp(enc_pass, row[1]) != 0) {
	    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authenticate): Wrong password for user `%{SQL-User-Name}' [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
	    snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
	    radlog(L_AUTH, querystr);
	    DEBUG2("rlm_nibs (rlm_nibs_authenticate): request_passwd = `%s'; sql_pass = `%s'; crypt_pass = `%s'", request->password->strvalue, row[1], enc_pass);
	    /* Remove the username we (maybe) added above */
	    (inst->module->sql_finish_select_query)(sqlsocket, inst->config);
            sql_release_socket(inst, sqlsocket);
	    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	    if(enc_pass != NULL) free(enc_pass);
	    return RLM_MODULE_REJECT;
	}
	if(enc_pass != NULL) free(enc_pass);
	(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
        sql_release_socket(inst, sqlsocket);
	pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	return RLM_MODULE_OK;
}

static int rlm_nibs_postauth(void *instance, REQUEST * request)
{
	SQL_INST	*inst = instance;
	NIBS_USER	*nibs_user;
	int		nibs_errno;
	char		*nibs_error;
	VALUE_PAIR	*vp;
	char		querystr[MAX_QUERY_LEN];
	char		querystr2[MAX_QUERY_LEN];
	char		*call_from;
	/* sqlusername holds the sql escaped username. The original
	 * username is at most MAX_STRING_LEN chars long and
	 * *sql_escape_string doubles its length in the worst case.
	 * Throw in an extra 10 to account for trailing NULs and to have
	 * a safety margin. */
	char   sqlusername[2 * MAX_STRING_LEN + 10];

        if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL) {
	    call_from = vp->strvalue;
	} else {
	    call_from = NULL;
	}

	/*
	 *	They MUST have a user name to do post_auth.
	 */
	if ((request->username == NULL) ||
	    (request->username->length == 0)) {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_postauth): Zero length username not permitted [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_ERR, querystr);
		return RLM_MODULE_INVALID;
	}

	/*
	 * Set, escape, and check the user attr here
	 */
	if (sql_set_user(inst, request, sqlusername, NULL) < 0)
		return RLM_MODULE_FAIL;
	
//FILTER
	if((nibs_user = nibs_fill_user(inst, request)) != NULL) {
	    if ((nibs_errno = nibs_filter(nibs_user, inst, request)) != 0) {
		nibs_error = nibs_decode_filter_error(nibs_errno, inst, request);
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_authenticate): User `%{SQL-User-Name}' login failed%%s%%s%%s [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (nibs_error != NULL) ? " (" : "", (nibs_error != NULL) ? nibs_error : "", (nibs_error != NULL) ? ")" : "", (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_AUTH, querystr);
		if (nibs_error != NULL) free(nibs_error);
		free(nibs_user);
		pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
		return RLM_MODULE_REJECT;
	    }
	} else {
		pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
		return RLM_MODULE_NOTFOUND;
	}

	radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_postauth): User `%{SQL-User-Name}' login OK [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
	snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
	radlog(L_AUTH, querystr);
	free(nibs_user);
	pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
	return RLM_MODULE_OK;
}

/*
 *	Accounting: save the account data to our sql table
 */
static int rlm_nibs_accounting(void *instance, REQUEST * request) {

	SQLSOCK *sqlsocket = NULL;
	VALUE_PAIR *pair;
	SQL_INST *inst = instance;
	int	ret = RLM_MODULE_OK;
//	int     numaffected = 0;
	int     acctstatustype = 0;
	int     record_found = 0;
	char    querystr[MAX_QUERY_LEN];
	char    querystr2[MAX_QUERY_LEN];
	char    logstr[MAX_QUERY_LEN];
	char	sqlusername[2 * MAX_STRING_LEN + 10];
	NIBS_USER *nibs_user;
	int	nibs_errno;
	char	*nibs_error;
	char	*call_from;

/*
#ifdef CISCO_ACCOUNTING_HACK
	int     acctsessiontime = 0;
#endif
*/
	memset(querystr, 0, MAX_QUERY_LEN);
	memset(querystr2, 0, MAX_QUERY_LEN);

        if ((pair = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL) {
	    call_from = pair->strvalue;
	} else {
	    call_from = NULL;
	}

	/*
	 *	They MUST have a user name and password to do NIBS authorize.
	 */

	if ((request->username == NULL) || (request->username->length == 0)) {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_accounting): Zero length username not permitted [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
		radlog(L_ERR, querystr);
		return RLM_MODULE_INVALID;
	}

	/*
	 * Find the Acct Status Type
	 */
	if ((pair = pairfind(request->packet->vps, PW_ACCT_STATUS_TYPE)) != NULL) {
		acctstatustype = pair->lvalue;
	} else {
		radius_xlat(logstr, sizeof(logstr), "rlm_nibs (rlm_nibs_accounting): packet has no account status type.  [user '%{SQL-User-Name}', nas '%{NAS-IP-Address}']", request, sql_escape_func);
		radlog(L_ERR, logstr);
		return RLM_MODULE_INVALID;
	}

	switch (acctstatustype) {
			/*
			 * The Terminal server informed us that it was rebooted
			 * STOP all records from this NAS 
			 */
/* FIXME
		case PW_STATUS_ACCOUNTING_ON:
		case PW_STATUS_ACCOUNTING_OFF:
			radlog(L_INFO, "rlm_nibs (%s): received Acct On/Off packet", inst->config->xlat_name);
			radius_xlat(querystr, sizeof(querystr), inst->config->accounting_onoff_query, request, sql_escape_func);
			query_log(request, inst, querystr);

			sqlsocket = sql_get_socket(inst);
			if (sqlsocket == NULL)
				return(RLM_MODULE_FAIL);
			if (*querystr) {
				if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					radlog(L_ERR, "rlm_nibs (%s): Couldn't update SQL accounting for Acct On/Off packet - %s",
					       inst->config->xlat_name,
					       (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					ret = RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
			}

			break;

*/

			/*
			 * Got accounting start packet
			 */
		case PW_STATUS_START:

			/*
			 * Set, escape, and check the user attr here
			 */
			sql_set_user(inst, request, sqlusername, NULL);

			if((nibs_user = nibs_fill_user(inst, request)) != NULL) {
			    if (nibs_user->no_acct != 0) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_OK;
			    }

			    if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_FAIL;
			    }

			    if(record_found > 0) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_OK;
			    }

			    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_start_query, request, sql_escape_func);

			    if (*querystr2) { /* non-empty query */
				sqlsocket = sql_get_socket(inst);

				if (sqlsocket == NULL) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				snprintf(querystr, sizeof(querystr), querystr2, nibs_user->gid, nibs_user->deposit);
			    	query_log(request, inst, querystr);

				if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_start_query' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					return RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
				sql_release_socket(inst, sqlsocket);
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_start_query' configured");
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				sql_release_socket(inst, sqlsocket);
				return RLM_MODULE_FAIL;
			    }

			    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
			    free(nibs_user);
			    return RLM_MODULE_OK;
			}

			return RLM_MODULE_NOOP;
			break;

			/*
			 * Got an update accounting packet
			 */
		case PW_STATUS_ALIVE:
			/*
			 * Set, escape, and check the user attr here
			 */
			sql_set_user(inst, request, sqlusername, NULL);

			if((nibs_user = nibs_fill_user(inst, request)) != NULL) {
			    if (nibs_user->no_acct != 0) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_OK;
			    }

			    nibs_add_current(nibs_user, inst, request);

			    if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_FAIL;
			    }

			    /* ALIVE record without START */
			    if(record_found == 0) {
				/*
			    	    Maybe STOP already found?
			    	    Try to reopen it!
			    	*/
				radius_xlat(querystr, sizeof(querystr), inst->config->accounting_reopen, request, sql_escape_func);

				if (*querystr) { /* non-empty query */
				    sqlsocket = sql_get_socket(inst);

				    if (sqlsocket == NULL) {
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					return RLM_MODULE_FAIL;
				    }

			    	    query_log(request, inst, querystr);

				    if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_reopen' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					sql_release_socket(inst, sqlsocket);
					return RLM_MODULE_FAIL;
				    }
				    (inst->module->sql_finish_query)(sqlsocket, inst->config);
				    sql_release_socket(inst, sqlsocket);
				} else {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_reopen' configured");
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}

				if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				if(record_found == 0) {
				    /* STOP not found too, so add START*/
				    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_start_query_alt, request, sql_escape_func);

				    if (*querystr2) { /* non-empty query */
					sqlsocket = sql_get_socket(inst);

					if (sqlsocket == NULL) {
					    free(nibs_user);
					    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					    return RLM_MODULE_FAIL;
					}

					snprintf(querystr, sizeof(querystr), querystr2, nibs_user->gid, nibs_user->current_time_on, nibs_user->deposit);
			    		query_log(request, inst, querystr);

					if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_start_query_alt' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					    free(nibs_user);
					    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					    sql_release_socket(inst, sqlsocket);
					    return RLM_MODULE_FAIL;
					}
					(inst->module->sql_finish_query)(sqlsocket, inst->config);
					sql_release_socket(inst, sqlsocket);
				    } else {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_start_query_alt' configured");
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					sql_release_socket(inst, sqlsocket);
					return RLM_MODULE_FAIL;
				    }

				    if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					return RLM_MODULE_FAIL;
				    }
				    radlog(L_INFO, "rlm_nibs (rlm_nibs_accounting): ALIVE without START for user `%s', so added START", request->username->strvalue);
				} else {
				    radlog(L_INFO, "rlm_nibs (rlm_nibs_accounting): Session for user `%s' reopened by ALIVE", request->username->strvalue);
				}
			    }

			    if((nibs_user->time_on == nibs_user->current_time_on) &&
			       (nibs_user->in_bytes == nibs_user->current_in_bytes) &&
			       (nibs_user->out_bytes == nibs_user->current_out_bytes)) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_OK;
			    }
	
			    nibs_billing(nibs_user, inst, request, acctstatustype);
	
			    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_alive_auth_query, request, sql_escape_func);
			    if (*querystr2) {
				sqlsocket = sql_get_socket(inst);

				if (sqlsocket == NULL) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				snprintf(querystr, sizeof(querystr), querystr2, ((nibs_user->do_with_tos == 1) ? nibs_user->sub_money : 0), nibs_user->sub_total_traffic, nibs_user->sub_total_time, nibs_user->sub_total_money);
				query_log(request, inst, querystr);

			    	if(rlm_nibs_query(sqlsocket, inst, querystr)) {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_alive_auth_query' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
				sql_release_socket(inst, sqlsocket);
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_alive_auth_query' configured");
			    }

			    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_update_query, request, sql_escape_func);
			    if (*querystr2) {
				sqlsocket = sql_get_socket(inst);

				if (sqlsocket == NULL) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				snprintf(querystr, sizeof(querystr), querystr2, nibs_user->sub_money, nibs_user->current_hour_in_bytes, nibs_user->current_hour_out_bytes, nibs_user->current_hour_traffic_money);
				query_log(request, inst, querystr);

				if(rlm_nibs_query(sqlsocket,inst,querystr)) {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_update_query' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
				sql_release_socket(inst, sqlsocket);
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_update_query' configured");
			    }

			    if(inst->config->kill_by_alive == 1) {
				nibs_user->deposit -= nibs_user->sub_money;
				nibs_user->total_time += nibs_user->sub_total_time;
				nibs_user->total_traffic += nibs_user->sub_total_traffic;

				if ((nibs_errno = nibs_filter(nibs_user, inst, request)) != 0) {
				    nibs_error = nibs_decode_filter_error(nibs_errno, inst, request);
				    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_accounting): User `%{SQL-User-Name}' must be killed%%s%%s%%s [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
				    snprintf(querystr, sizeof(querystr), querystr2, (nibs_error != NULL) ? " (" : "", (nibs_error != NULL) ? nibs_error : "", (nibs_error != NULL) ? ")" : "", (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
				    radlog(L_INFO, querystr);
				    if (nibs_error != NULL) free(nibs_error);
				    nibs_user_kill(inst, request);
				}
			    }

			    free(nibs_user);
			    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
			    return RLM_MODULE_OK;
			}

			return RLM_MODULE_NOOP;
			break;

			/*
			 * Got accounting stop packet
			 */
		case PW_STATUS_STOP:

			/*
			 * Set, escape, and check the user attr here
			 */
			sql_set_user(inst, request, sqlusername, NULL);

			if((nibs_user = nibs_fill_user(inst, request)) != NULL) {
			    if (nibs_user->no_acct != 0) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_OK;
			    }

			    nibs_add_current(nibs_user, inst, request);

			    if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
				free(nibs_user);
				pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				return RLM_MODULE_FAIL;
			    }

			    /* STOP record without START*/
			    if(record_found == 0) {
				/*
			    	    Maybe STOP already found?
			    	    Try to reopen it!
			    	*/
				radius_xlat(querystr, sizeof(querystr), inst->config->accounting_reopen, request, sql_escape_func);

				if (*querystr) { /* non-empty query */
				    sqlsocket = sql_get_socket(inst);

				    if (sqlsocket == NULL) {
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					return RLM_MODULE_FAIL;
				    }

			    	    query_log(request, inst, querystr);

				    if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_reopen' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					sql_release_socket(inst, sqlsocket);
					return RLM_MODULE_FAIL;
				    }
				    (inst->module->sql_finish_query)(sqlsocket, inst->config);
				    sql_release_socket(inst, sqlsocket);
				} else {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_reopen' configured");
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}

				if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}


				if(record_found == 0) {
				    /* STOP not found too, so add START*/
				    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_start_query_alt, request, sql_escape_func);

				    if (*querystr2) { /* non-empty query */
					sqlsocket = sql_get_socket(inst);

					if (sqlsocket == NULL) {
					    free(nibs_user);
					    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					    return RLM_MODULE_FAIL;
					}

					snprintf(querystr, sizeof(querystr), querystr2, nibs_user->gid, nibs_user->current_time_on, nibs_user->deposit);
			    		query_log(request, inst, querystr);

					if (rlm_nibs_query(sqlsocket, inst, querystr)) {
					    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_start_query_alt' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
					    free(nibs_user);
					    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					    sql_release_socket(inst, sqlsocket);
					    return RLM_MODULE_FAIL;
					}

					(inst->module->sql_finish_query)(sqlsocket, inst->config);
					sql_release_socket(inst, sqlsocket);
				    } else {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_start_query_alt' configured");
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					sql_release_socket(inst, sqlsocket);
					return RLM_MODULE_FAIL;
				    }

				    if((record_found = nibs_user_add_acct(nibs_user, inst, request)) == -1) {
					free(nibs_user);
					pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
					return RLM_MODULE_FAIL;
				    }
				    radlog(L_INFO, "rlm_nibs (rlm_nibs_accounting): STOP without START for user `%s', so added START", request->username->strvalue);
				} else {
				    radlog(L_INFO, "rlm_nibs (rlm_nibs_accounting): Session for user `%s' reopened by STOP", request->username->strvalue);
				}
			    }

			    nibs_billing(nibs_user, inst, request, acctstatustype);

			    if ((pair = pairfind(request->packet->vps, PW_ACCT_TERMINATE_CAUSE)) != NULL) {
				if (pair->lvalue == 0) {
				    pairdelete(&request->packet->vps, PW_ACCT_TERMINATE_CAUSE);
				    pair = NULL;
				}
			    }

			    if(pair == NULL) {
				if ((pair = paircreate(PW_ACCT_TERMINATE_CAUSE, PW_TYPE_INTEGER)) == NULL) {
				    radlog(L_ERR, "%s", librad_errstr);
				} else {
				    pair->lvalue = 1;
		    		    pair->operator = T_OP_EQ;
    				    pairadd(&request->packet->vps, pair);
				}
			    }

			    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_stop_auth_query, request, sql_escape_func);
			    if (*querystr2) {
				sqlsocket = sql_get_socket(inst);

				if (sqlsocket == NULL) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				snprintf(querystr, sizeof(querystr), querystr2, ((nibs_user->do_with_tos == 1) ? nibs_user->sub_money : 0), nibs_user->sub_total_traffic, nibs_user->sub_total_time, nibs_user->sub_total_money);
				query_log(request, inst, querystr);

			    	if(rlm_nibs_query(sqlsocket, inst, querystr)) {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_stop_auth_query' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
				sql_release_socket(inst, sqlsocket);
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_stop_auth_query' configured");
			    }

			    radius_xlat(querystr2, sizeof(querystr2), inst->config->accounting_stop_query, request, sql_escape_func);
			    if (*querystr2) {
				sqlsocket = sql_get_socket(inst);

				if (sqlsocket == NULL) {
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    return RLM_MODULE_FAIL;
				}

				snprintf(querystr, sizeof(querystr), querystr2, nibs_user->sub_money);
				query_log(request, inst, querystr);

				if(rlm_nibs_query(sqlsocket,inst,querystr)) {
				    radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): Couldn't finish `accounting_stop_query' for user `%s' - database error `%s'", request->username->strvalue, (char *)(inst->module->sql_error)(sqlsocket, inst->config));
				    free(nibs_user);
				    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
				    sql_release_socket(inst, sqlsocket);
				    return RLM_MODULE_FAIL;
				}
				(inst->module->sql_finish_query)(sqlsocket, inst->config);
				sql_release_socket(inst, sqlsocket);
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_accounting): No `accounting_stop_query' configured");
			    }

			    free(nibs_user);
			    radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_accounting): User `%{SQL-User-Name}' logout [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
			    snprintf(querystr, sizeof(querystr), querystr2, (call_from) ? " |" : "", (call_from) ? call_from : "", (call_from) ? "|" : "");
			    radlog(L_AUTH, querystr);
			    pairdelete(&request->packet->vps, PW_SQL_USER_NAME);
			    return RLM_MODULE_OK;
			}

			return RLM_MODULE_NOOP;
			break;

			/*
			 *	Anything else is ignored.
			 */
		default:
			radlog(L_INFO, "rlm_nibs (rlm_nibs_accounting): Unsupported Acct-Status-Type = %d", acctstatustype);
			return RLM_MODULE_NOOP;
			break;

	}

	return ret;
}

/*
 *        See if a user is already logged in. Sets request->simul_count to the
 *        current session count for this user.
 * 
 *        Check twice. If on the first pass the user exceeds his
 *        max. number of logins, do a second pass and validate all
 *        logins by querying the terminal server (using eg. SNMP).
 */

static int rlm_nibs_checksimul(void *instance, REQUEST * request) {
	SQLSOCK 	*sqlsocket;
	SQL_INST	*inst = instance;
	SQL_ROW		row;
	char		querystr[MAX_QUERY_LEN];
	char		querystr2[MAX_QUERY_LEN];
	char		sqlusername[2*MAX_STRING_LEN+10];
	int		check = 0;
        uint32_t        ipno = 0;
        char            *call_num = NULL;
	VALUE_PAIR      *vp;
	LRAD_TOKEN	parsecode;
	VALUE_PAIR      *pair;
	uint32_t	nas_addr = 0;
	int		nas_port = 0;
	uint32_t	request_nas_addr = 0;
	int		request_nas_port = 0;
	NIBS_USER	*nibs_user = NULL;
	int		acct_interval = 0;

        if ((vp = pairfind(request->packet->vps, PW_NAS_IP_ADDRESS)) != NULL)
		request_nas_addr = inet_addr(vp->strvalue);
        if ((vp = pairfind(request->packet->vps, PW_NAS_PORT)) != NULL)
                request_nas_port = vp->lvalue;

        if ((vp = pairfind(request->packet->vps, PW_CALLING_STATION_ID)) != NULL)
                call_num = vp->strvalue;

	if ((request->username == NULL) || (request->username->length == 0)) {
		radius_xlat(querystr2, sizeof(querystr2), "rlm_nibs (rlm_nibs_checksimul): Zero length username not permitted [%{NAS-IP-Address}:%{NAS-Port}]%%s%%s%%s", request, sql_escape_func);
		snprintf(querystr, sizeof(querystr), querystr2, (call_num) ? " |" : "", (call_num) ? call_num : "", (call_num) ? "|" : "");
		radlog(L_ERR, querystr);
		return RLM_MODULE_INVALID;
	}

	if(sql_set_user(inst, request, sqlusername, 0) < 0)
		return RLM_MODULE_FAIL;

/* At first cheacking for stalesessions for ALL*/
	if (inst->config->simul_verify_query[0] == 0) {
		radlog(L_INFO,"rlm_nibs (rlm_nibs_checksimul): No `simul_verify_query' configured in nibs.conf file");
		return RLM_MODULE_NOOP;
	}

	radius_xlat(querystr, sizeof(querystr), inst->config->simul_verify_query, request, NULL);

	/* initialize the sql socket */
	sqlsocket = sql_get_socket(inst);
	if(sqlsocket == NULL)
		return RLM_MODULE_FAIL;

	if(rlm_nibs_select_query(sqlsocket, inst, querystr)) {
		radlog(L_ERR, "rlm_nibs (rlm_nibs_checksimul): Database query error 1");
		sql_release_socket(inst, sqlsocket);
		return RLM_MODULE_FAIL;
	}

	while (rlm_nibs_fetch_row(sqlsocket, inst) == 0) {
		row = sqlsocket->row;
		if (row == NULL)
			break;
		if (!row[1]){
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul): Cannot zap stale entry. No username present in entry.");
			return RLM_MODULE_FAIL;
		}
		if (!row[0]){
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul): Cannot zap stale entry. No session id in entry.");
			return RLM_MODULE_FAIL;
		}
		if (row[2])
			nas_addr = inet_addr(row[2]);
		if (row[3])
			nas_port = Atoi(row[3]);

/* check for port */
		if ((nas_addr == request_nas_addr) && (nas_port == request_nas_port)) {
                    	    /*
			     *      Stale record - zap it.
                	     */
			uint32_t framed_addr = 0;
			char proto = 'P';
			if (row[4])
			    framed_addr = inet_addr(row[4]);
			if (row[5])
			    if (strcmp(row[5],"SLIP") == 0)
			        proto = 'S';

			radlog(L_INFO, "rlm_nibs (rlm_nibs_checksimul): User `%s' zapped by port eq", row[1]);
			session_zap(request->packet->sockfd, nas_addr,nas_port,row[1],row[0], framed_addr, proto,0);
			continue;
		}

/* check by ALIVE */
		if (inst->config->deletestalesessions_by_alive) {
		    if ((nibs_user == NULL) || (strcmp(nibs_user->name, row[1]) != 0)) {
			if (nibs_user) free(nibs_user);
			if ((nibs_user = nibs_fill_user(inst, request)) != NULL) {
			    if (nibs_user->other_params[0] != 0) {
				if (nibs_user->other_params[strlen(nibs_user->other_params) - 1] == ',') {
    				    nibs_user->other_params[strlen(nibs_user->other_params) - 1] = '\0';
				}
				pair = NULL;
				parsecode = userparse(nibs_user->other_params, &pair);

				if (parsecode < 1) {
				    radlog(L_ERR|L_CONS, "rlm_nibs (rlm_nibs_checksimul): Parse error (reply) at SQL other_params");
				} else {
    				    if ((vp = pairfind(pair, PW_ACCT_INTERIM_INTERVAL)) != NULL) {
            				acct_interval = vp->lvalue;
				    } else {
					radlog(L_ERR, "rlm_nibs (rlm_nibs_checksimul): No Acct-Interim-Interval configured in SQL other_params");
				    }
				    pairfree(&pair);
				}
			    } else {
				radlog(L_ERR, "rlm_nibs (rlm_nibs_checksimul): No SQL other_params configured");
			    }
			} else {
			    DEBUG("rlm_nibs (rlm_nibs_checksimul): Cannot fill user.");
			}
		    }

		    if ((request->timestamp - Atoi(row[6])) >= (acct_interval + inst->config->alive_max_interval)) {
                    	    /*
			     *      Stale record - zap it.
                	     */
			uint32_t framed_addr = 0;
			char proto = 'P';
			if (row[4])
			    framed_addr = inet_addr(row[4]);
			if (row[5])
			    if (strcmp(row[5],"SLIP") == 0)
			        proto = 'S';

			radlog(L_INFO, "rlm_nibs (rlm_nibs_checksimul): User `%s' zapped by max ALIVE counter ((Now:%d - Last_ALIVE:%d) >= (Acct-Interim:%d + Max_Interval:%d))", row[1], (int)request->timestamp, Atoi(row[6]), (int)acct_interval, (int)inst->config->alive_max_interval);
			session_zap(request->packet->sockfd, nas_addr,nas_port,row[1],row[0], framed_addr, proto,0);
			continue;
		    }
		} // deletestalesessions_by_alive

/* check by checkrad.pl */
		if (inst->config->deletestalesessions_by_checkrad) {

		    check = rad_check_ts(nas_addr, nas_port, row[1], row[0]);

            	    /*
                     *      Failed to check the terminal server for
                     *      duplicate logins: Return an error.
                     */
		    if (check < 0) {
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul) rad_check_ts() failed.");
			return RLM_MODULE_FAIL;
		    }

		    if(check != 1) {
                        /*
                         *      Stale record - zap it.
                         */
			uint32_t framed_addr = 0;
			char proto = 'P';

			if (row[4])
				framed_addr = inet_addr(row[4]);
			if (row[5])
				if (strcmp(row[5],"SLIP") == 0)
					proto = 'S';

			radlog(L_INFO, "rlm_nibs (rlm_nibs_checksimul): User `%s' zapped by checkrad script", row[1]);
			session_zap(request->packet->sockfd, nas_addr,nas_port,row[1],row[0], framed_addr, proto,0);
		    }
		}
	}// select
	if (nibs_user) free(nibs_user);
	(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
	sql_release_socket(inst, sqlsocket);


/* Second count Online sessions for cheacked user */

	/* Looks like too many sessions, so lets start verifying them */

	/* If simul_count_query is not defined, we don't do any checking */
	if (inst->config->simul_count_query[0] == 0) {
		radlog(L_INFO,"rlm_nibs (rlm_nibs_checksimul): No `simul_count_query' configured in nibs.conf file");
		return RLM_MODULE_NOOP;
	}

	radius_xlat(querystr, sizeof(querystr), inst->config->simul_count_query, request, NULL);

	/* initialize the sql socket */
	sqlsocket = sql_get_socket(inst);
	if(sqlsocket == NULL)
		return RLM_MODULE_FAIL;

	if(rlm_nibs_select_query(sqlsocket, inst, querystr)) {
		radlog(L_ERR, "rlm_nibs (rlm_nibs_checksimul): Database query error 2");
		sql_release_socket(inst, sqlsocket);
		return RLM_MODULE_FAIL;
	}

	request->simul_count = (inst->module->sql_num_rows)(sqlsocket, inst->config);

	if(request->simul_count < request->simul_max) {
		sql_release_socket(inst, sqlsocket);
		return RLM_MODULE_OK;
	}

        /*
         *      Setup some stuff, like for MPP detection.
         */

	request->simul_count = 0;

        if ((vp = pairfind(request->packet->vps, PW_FRAMED_IP_ADDRESS)) != NULL)
                ipno = vp->lvalue;

	while (rlm_nibs_fetch_row(sqlsocket, inst) == 0) {
		row = sqlsocket->row;
		if (row == NULL)
			break;
		if (!row[1]){
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul): Cannot zap stale entry. No username present in entry.");
			return RLM_MODULE_FAIL;
		}
		if (!row[0]){
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul): Cannot zap stale entry. No session id in entry.");
			return RLM_MODULE_FAIL;
		}
		if (row[2])
			nas_addr = inet_addr(row[2]);
		if (row[3])
			nas_port = Atoi(row[3]);

		check = rad_check_ts(nas_addr, nas_port, row[1], row[0]);

            	    /*
                    *      Failed to check the terminal server for
                    *      duplicate logins: Return an error.
                    */
		if (check < 0) {
			(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
			sql_release_socket(inst, sqlsocket);
			DEBUG("rlm_nibs (rlm_nibs_checksimul) rad_check_ts() failed.");
			return RLM_MODULE_FAIL;
		}

		if(check == 1) {
			++request->simul_count;

                        /*
                         *      Does it look like a MPP attempt?
                         */
                        if (row[4] && ipno && inet_addr(row[4]) == ipno)
                                request->simul_mpp = 2;
                        else if (row[5] && call_num &&
                                !strncmp(row[5],call_num,16))
                                request->simul_mpp = 2;
		}
	}
	(inst->module->sql_finish_select_query)(sqlsocket, inst->config);
	sql_release_socket(inst, sqlsocket);

	/* The Auth module apparently looks at request->simul_count, not the return value
	   of this module when deciding to deny a call for too many sessions */

	return RLM_MODULE_OK;

}

/* globally exported name */
module_t rlm_nibs = {
	"NIBS",
	RLM_TYPE_THREAD_SAFE,	/* type: reserved */
	rlm_nibs_init,		/* initialization */
	rlm_nibs_instantiate,	/* instantiation */
	{
		rlm_nibs_authenticate,	/* authenticate */
		rlm_nibs_authorize,	/* authorize */
		NULL,			/* preaccounting */
		rlm_nibs_accounting,	/* accounting */
		rlm_nibs_checksimul,	/* checksimul */
		NULL,			/* pre-proxy */
		NULL,			/* post-proxy */
		rlm_nibs_postauth,	/* post-auth */
	},
	rlm_nibs_detach,		/* detach */
	rlm_nibs_destroy,	/* destroy */
};
